/*
Copyright (C) 2001 Charles Samuels <charles@kde.org>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public License
along with this library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*/

#ifndef __ITC_PIPE_H
#define __ITC_PIPE_H

#include <itclib/thread.h>
#include <string.h>

namespace ITC
{

class GPipe
{
public:
	GPipe(int size=16384);
	~GPipe();

	int read(int length, void *data, bool block=true);

	/**
	 * calls the above read with force=false, then
	 * fills all remaining space with the given byte
	 **/
	int read(int length, char fill, void *data, bool block=false);

//	int peek(int length, void *data, bool block=true) const; // TODO
	int write(int length, void *data, bool block=true);

	
	int size() const;
	int free() const;
 
	void flush();
	void flush(int size);
 
	void close();
	bool closed(bool only) const { return mClosed && (only || !size()); }

private:
	mutable ITC::Mutex mMutexStart, mMutexEnd, mMutexRead, mMutexWrite;
	mutable ITC::Condition mCondStart, mCondEnd;
	
	void *buffer;
	void *bufferEnd;

	void *start;
	void *end;

	bool mClosed;
};

/**
 * A mutexed and safe circular buffer.  It's like a high-speed
 * FIFO, and highly optimized for speed.  It can also be much
 * larger than a kernel FIFO (a pipe)
 **/
template<class Type> class Pipe : protected GPipe
{
public:
	Pipe(int count=16384) : GPipe(count*sizeof(Type))
	{}
 
	/**
	 * return the number of items read
	 *
	 * if block is true, as it is by default, then
	 * always fill the buffer 'data' entirely,
	 * and wait if necessary
	 *
	 * even if block is true, it can still return a number
	 * less than length, if EOF is reached
	**/
	int read(int length, Type *data, bool block=true)
		{ return GPipe::read(length*sizeof(Type), (void*)data, block)/sizeof(Type);}
  
	/**
	 * calls the above read with block=false, then
	 * fills all remaining values with the given value
	 **/
	int read(int length, Type fill, Type *data)
	{
		int len = GPipe::read(length*sizeof(Type), data, false);
		len /= sizeof(Type);
	
		for (; len<length; len++)
		{
			(((Type*)data)[len])=fill;
		}

		return len;
	}

	/**
	 * denote that read should never block, and
	 * always return -1 if the buffer is empty,
	 * or the amount of data returned
	 *
	 * unless you start appending again
	 **/
	void close()
		{ GPipe::close(); }

	/**
	 * check if the pipe is closed.
	 * if only is true, it'l check if the pipe has been closed
	 * on the other end, even if there's more data
	 **/
	bool closed(bool only=false) const
		{ return GPipe::closed(only); }
	/**
	 * and it will unclose the fifo if needed
	 * and return the amount unwritten,
	 * which will always be zero, unless block is false
	 **/
	int write(int length, Type *data, bool block=true)
		{ return GPipe::write(length*sizeof(Type), (void*)data, block)/sizeof(Type); }

	/**
	 * this is akin to reading all the data
	 **/
	void flush() { GPipe::flush(); }

	int size() const { return GPipe::size()/sizeof(Type); }
	int free() const { return GPipe::free()/sizeof(Type); }

	void flush(int size) { GPipe::flush(size*sizeof(Type)); }
};




}

#endif
